/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.release.bundle;

import lombok.NonNull;
import org.artifactory.api.component.ComponentDetails;
import org.artifactory.api.component.ComponentDetailsFetcher;
import org.artifactory.api.release.bundle.ReleaseBundleService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.rest.distribution.bundle.models.ReleaseArtifact;
import org.artifactory.api.rest.distribution.bundle.models.ReleaseBundleModel;
import org.artifactory.bundle.BundleTransactionStatus;
import org.artifactory.bundle.BundleType;
import org.artifactory.bundle.LightArtifactsBundle;
import org.artifactory.model.xstream.bundle.ArtifactsBundleImpl;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.storage.db.bundle.dao.ArtifactBundlesDao;
import org.artifactory.storage.db.bundle.dao.BundleBlobsDao;
import org.artifactory.storage.db.bundle.model.BundleNode;
import org.artifactory.storage.db.bundle.model.DBArtifactsBundle;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.entity.NodePath;
import org.jfrog.storage.StorageException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.artifactory.util.distribution.DistributionConstants.IN_TRANSIT_REPO_KEY;

/**
 * @author Rotem Kfir
 */
@Service
public class ReleaseBundleServiceImpl implements ReleaseBundleService {
    private static final Logger log = LoggerFactory.getLogger(ReleaseBundleServiceImpl.class);

    private ArtifactBundlesDao bundlesDao;

    private BundleBlobsDao blobsDao;

    private NodesDao nodesDao;

    private RepositoryService repoService;

    private ArtifactBundlesDao artifactsBundleDao;

    private ComponentDetailsFetcher componentDetailsFetcher;

    @Autowired
    public ReleaseBundleServiceImpl(ArtifactBundlesDao bundlesDao, BundleBlobsDao blobsDao,
                                    NodesDao nodesDao, RepositoryService repoService,
                                    ArtifactBundlesDao artifactsBundleDao,
                                    ComponentDetailsFetcher componentDetailsFetcher) {
        this.bundlesDao = bundlesDao;
        this.blobsDao = blobsDao;
        this.nodesDao = nodesDao;
        this.repoService = repoService;
        this.artifactsBundleDao = artifactsBundleDao;
        this.componentDetailsFetcher = componentDetailsFetcher;
    }

    @Override
    @Nonnull
    public List<Long> findExpiredBundlesIds(Long cleanupPeriod) {
        try {
            return bundlesDao.getExpiredBundlesIds(cleanupPeriod);
        } catch (SQLException e) {
            throw new StorageException("Could not get expired Release Bundle transactions", e);
        }
    }

    @Override
    public boolean deleteBundle(Long id, boolean deleteContent) {
        log.debug("Deleting Release Bundle with id '{}'", id);
        try {
            DBArtifactsBundle artifactsBundle = bundlesDao.getArtifactsBundle(id);
            if (artifactsBundle == null) {
                return false;
            }
            List<BundleNode> bundleNodes = bundlesDao.getBundleNodes(artifactsBundle.getId());
            if (!bundleNodes.isEmpty()) {
                bundlesDao.deleteBundleNodes(bundleNodes.stream()
                        .map(BundleNode::getId)
                        .collect(Collectors.toList()));
            }
            if (deleteContent) {
                deleteBundleContent(artifactsBundle, bundleNodes);
            }
            blobsDao.deleteBlob(artifactsBundle.getId());
            bundlesDao.deleteArtifactsBundle(artifactsBundle.getId());
            return true;
        } catch (SQLException e) {
            throw new StorageException("Could not delete Release Bundle", e);
        }
    }

    @Override
    public boolean redistributeBundle(Long id) {
        log.info("Redistributing Release Bundle with id '{}'", id);
        try {
            return bundlesDao.updateArtifactsBundleCreationDate(id) > 0;
        } catch (SQLException e) {
            throw new StorageException("Could not redistribute Release Bundle with id '" + id + "'", e);
        }
    }

    @Override
    @Nonnull
    public List<LightArtifactsBundle> getBundlesRelatedToRepoPath(String path) {
        try {
            List<DBArtifactsBundle> artifactsBundles = bundlesDao.getBundlesRelatedToRepoPath(path);
            return artifactsBundles.stream()
                    .map(this::dbBundleToLightBundle)
                    .collect(Collectors.toList());
        } catch (SQLException e) {
            throw new StorageException("Could not check if Repo path '" + path + "' is related to a Release Bundle", e);
        }
    }

    private LightArtifactsBundle dbBundleToLightBundle(DBArtifactsBundle dbArtifactsBundle) {
        return ArtifactsBundleImpl.builder()
                .id(dbArtifactsBundle.getId())
                .name(dbArtifactsBundle.getName())
                .version(dbArtifactsBundle.getVersion())
                .build();
    }

    @Override
    public void deleteAllBundles() {
        try {
            bundlesDao.deleteAllBundleNodes();
            blobsDao.deleteAllBlobs();
            artifactsBundleDao.deleteAllArtifactsBundles();
        } catch (SQLException e) {
            log.error("Failed to delete all release bundles", e);
        }
    }

    @Override
    public void copyBundleArtifactsAsync(ReleaseBundleModel bundleModel, String storingRepo,
                                         Map<String, String> artifactMapping) throws SQLException {
        String prefix = bundleModel.getName() + "/" + bundleModel.getVersion();
        // artifacts should be stored as storingRepo/(RB-name/RB-version)/targetRepo/artifacts
        log.debug("Starting Async release bundle copy of {} to repository {}", prefix, storingRepo);
        for (ReleaseArtifact artifact : bundleModel.getArtifacts()) {
            try {
                copyArtifact(artifact, storingRepo, prefix, artifactMapping, bundleModel);
            } catch (Exception e) {
                log.error("Failed to copy artifact {} in source release bundle {}: {}", artifact.getRepoPath(), prefix,
                        e.getMessage());
                log.debug("", e);
                failBundle(bundleModel, prefix);
                return;
            }
        }
        // set bundle as completed
        int updated = bundlesDao
                .completeBundle(bundleModel.getName(), bundleModel.getVersion(), BundleType.SOURCE.name());
        if (updated == 0) {
            log.warn("Could not set source release bundle for {} as COMPLETED.", prefix);
        }
    }

    @Override
    public void createBundleNode(long artifactsBundleId, RepoPath targetRepoPath, ComponentDetails componentDetails) throws SQLException {
        long nodeId = nodesDao.getNodeId(NodePath.fromRepoPath(targetRepoPath));
        BundleNode bundleNode = new BundleNode();
        bundleNode.setBundleId(artifactsBundleId);
        bundleNode.setNodeId(nodeId);
        bundleNode.setRepoPath(targetRepoPath.toPath());
        if (componentDetails != null) { // in case this node is created at the target artifactory it doesn't have details
            bundleNode.setOriginalFileDetails(ComponentDetails.toJson(componentDetails));
        }
        artifactsBundleDao.create(bundleNode);
    }

    @Override
    public ComponentDetails getOriginalComponentDetails(@NonNull RepoPath repoPath) throws SQLException {
        String jsonDetails = artifactsBundleDao.getComponentDetails(repoPath.toPath());
        if (jsonDetails == null) {
            log.debug("Original component details does not exist for repoPath {}", repoPath);
            return null;
        }
        return ComponentDetails.fromJson(jsonDetails);
    }

    private void copyArtifact(ReleaseArtifact artifact, String storingRepo, String bundlePrefix,
                              Map<String, String> artifactMapping, ReleaseBundleModel bundleModel) throws SQLException {
        // target path is storingRepo/bundleName/bundleVersion/artifactPath
        String targetPath = bundlePrefix + '/' + artifact.getRepoPath();
        RepoPath target = RepoPathFactory.create(storingRepo, targetPath);
        String mapped = artifactMapping.get(artifact.getRepoPath());
        if (mapped == null) {
            String path = target.toPath();
            log.error("No artifact mapping provided for {}, can't determine source artifact", path);
            throw new IllegalArgumentException(
                    "No artifact mapping provided for " + path + ", can't determine source artifact");
        }
        RepoPath source = RepoPathFactory.create(mapped);
        log.trace("Copying {} to {}", source.getPath(), target.getPath());
        repoService.copy(source, target, false, true, false);
        try {
            ComponentDetails originalComponentDetails = componentDetailsFetcher.calcComponentDetails(source);
            createBundleNode(bundleModel.getBundleId(), target, originalComponentDetails);
        } catch (SQLException e) {
            log.error("Unable to add artifact '{}' to bundle {}", target.toPath(), bundlePrefix);
            throw e;
        }
    }

    private void failBundle(ReleaseBundleModel bundleModel, String prefix) throws SQLException {
        int updated = bundlesDao.failBundle(bundleModel.getName(), bundleModel.getVersion(), BundleType.SOURCE.name());
        if (updated == 0) {
            log.warn("Could not set source release bundle for {} as FAILED.", prefix);
        }
    }

    private void deleteBundleContent(DBArtifactsBundle artifactsBundle, List<BundleNode> bundleNodes)
            throws SQLException {
        if (BundleTransactionStatus.INPROGRESS.equals(artifactsBundle.getStatus())) {
            // Delete the temp folder and all of its content
            String txPath = String
                    .join("/", IN_TRANSIT_REPO_KEY, artifactsBundle.getName(), artifactsBundle.getVersion());
            repoService.undeploy(RepoPathFactory.create(txPath));
        } else {
            for (BundleNode bundleNode : bundleNodes) {
                // If the artifact is related to a different bundle, skip deletion
                if (bundlesDao.isNodeRelatedToBundle(bundleNode.getNodeId())) {
                    log.info("Artifact " + bundleNode.getRepoPath() +
                            " is related to a different release bundle and will not be deleted.");
                    continue;
                }
                repoService.undeploy(RepoPathFactory.create(bundleNode.getRepoPath()));
            }
        }
    }
}
