/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.service;

import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.api.repo.Async;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.checksum.ChecksumInfo;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.fs.ItemInfo;
import org.artifactory.fs.RepoResource;
import org.artifactory.repo.*;
import org.artifactory.repo.local.ValidDeployPathContext;
import org.artifactory.repo.service.versioning.RepositoriesCache;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.sapi.common.ExportSettings;
import org.artifactory.sapi.common.Lock;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.ContextReadinessListener;
import org.artifactory.spring.ReloadableBean;
import org.artifactory.storage.fs.MutableVfsItem;
import org.artifactory.storage.fs.service.ItemMetaInfo;
import org.artifactory.storage.jobs.migration.pathchecksum.RepoPathChecksumCalculationWorkItem;
import org.artifactory.storage.jobs.migration.sha256.ChecksumCalculationWorkItem;
import org.jfrog.common.config.diff.DataDiff;
import org.jfrog.storage.binstore.exceptions.BinaryRejectedException;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

/**
 * A repository service to be used by the core
 * <p/>
 * User: freds Date: Jul 31, 2008 Time: 5:50:18 PM
 */
public interface InternalRepositoryService extends RepositoryService, ReloadableBean, ContextReadinessListener,
        ContextCreationListener {

    boolean isAnonAccessEnabled();

    /**
     * @param key The repository key
     * @return Repository with the exact given key (no special meaning for remote/cache repo keys). Null if not found.
     */
    @Nullable
    Repo repositoryByKey(String key);

    @Nullable
    VirtualRepo virtualRepositoryByKey(String key);

    @Nullable
    LocalRepo localOrCachedRepositoryByKey(String key);

    /**
     * Get the holder object that holds the actual repository as well as the path.
     *
     * @param repoPath The repo path
     * @param <R>      The type of repository.
     * @return The holder object that holds the actual repository as well as the path.
     */
    <R extends Repo> RepoRepoPath<R> getRepoRepoPath(RepoPath repoPath);

    /**
     * Get a repository that can store artifacts. Will get either a local repo (regular or cache) or a virtual repo. If
     * remote repository key is given, the cache repository is returned.
     */
    StoringRepo storingRepositoryByKey(String key);

    List<RealRepo> getLocalAndRemoteRepositories();

    List<VirtualRepo> getVirtualRepositories();

    List<LocalRepo> getLocalAndCachedRepositories();

    /**
     * Returns a local non-cache repository by key
     *
     * @param key The repository key
     * @return Local non-cache repository or null if not found
     */
    LocalRepo localRepositoryByKey(String key);

    RemoteRepo remoteRepositoryByKey(String key);

    DistributionRepo distributionRepoByKey(String key);

    ReleaseBundlesRepo releaseBundleRepositoryByKey(String key);

    /**
     * This will verify the permission to deploy to the path
     *
     * @param validDeployPathContext Context for encapsulating all the methods parameters
     */
    void assertValidDeployPathAndPermissions(ValidDeployPathContext validDeployPathContext) throws RepoRejectException;

    <T extends RemoteRepoDescriptor> ResourceStreamHandle downloadAndSave(InternalRequestContext requestContext,
            RemoteRepo<T> remoteRepo, RepoResource res) throws IOException, RepoRejectException;

    @Lock
    RepoResource unexpireIfExists(LocalRepo localCacheRepo, String path);

    @Lock
    ResourceStreamHandle unexpireAndRetrieveIfExists(InternalRequestContext requestContext, LocalRepo localCacheRepo,
            String path) throws IOException, RepoRejectException;

    ResourceStreamHandle getResourceStreamHandle(InternalRequestContext requestContext, Repo repo, RepoResource res)
            throws IOException, RepoRejectException, BinaryRejectedException;

    /**
     * Saves the given resource in the input repository.
     *
     * @param repo        The repo to save the resource in. Must be a storage repo (local/local cache/virtual cache)
     * @param saveContext Resource details
     * @return Repository resource with details of the saved resource
     *
     * @throws IOException             On any IO related exception
     * @throws RepoRejectException     If the target repository rejected the resource
     * @throws BinaryRejectedException If the Binary failed validation at the storage layer
     */
    RepoResource saveResource(StoringRepo repo, SaveResourceContext saveContext)
            throws IOException, RepoRejectException, BinaryRejectedException;

    /**
     * Internal, transactional method to save a resource. Don't use directly.
     *
     * @see InternalRepositoryService#saveResource(org.artifactory.repo.StoringRepo, org.artifactory.repo.SaveResourceContext)
     */
    @Transactional
    RepoResource saveResourceInTransaction(StoringRepo repo, SaveResourceContext saveContext)
            throws IOException, RepoRejectException, BinaryRejectedException;

    @Override
    void exportTo(ExportSettings settings);

    boolean isRepoExistInCache(RepoPath repoPath);

    /**
     * Returns a local or local cache repository. Throws an exception if not found
     *
     * @param repoPath A repo path in the repository
     * @return Local/cache repository matching the repo path repo key
     *
     * @throws IllegalArgumentException if repository not found
     */
    @Nonnull
    LocalRepo getLocalRepository(RepoPath repoPath);

    @Override
    @Lock
    void reload(CentralConfigDescriptor oldDescriptor, List<DataDiff<?>> configDiff);

    /**
     * A convenient method for saving internally generated files to the repositories.
     *
     * @param fileRepoPath Repository path to store the input stream
     * @param is           The input stream to store. Will be closed before returning from this method
     * @see InternalRepositoryService#saveResource(org.artifactory.repo.StoringRepo, org.artifactory.repo.SaveResourceContext)
     */
    void saveFileInternal(RepoPath fileRepoPath, InputStream is) throws RepoRejectException, IOException;

    /**
     * @param itemInfo The item.
     * @return The meta info of the given repo path. Null if not found.
     */
    @Nullable
    ItemMetaInfo getItemMetaInfo(ItemInfo itemInfo);

    /**
     * Moves repository path (pointing to a folder) to another absolute target. The move will only move paths the user
     * has permissions to move and paths that are accepted by the target repository.
     *
     * @param from            The source path to move.
     * @param to              The target path to move to.
     * @param dryRun          If true the method will just report the expected result but will not move any file
     * @param suppressLayouts If true, path translation across different layouts should be suppressed.
     * @param failFast        If true, the operation should fail upon encountering an error.
     * @return MoveMultiStatusHolder holding the errors and warnings
     */
    @Lock
    MoveMultiStatusHolder moveWithoutMavenMetadata(RepoPath from, RepoPath to, boolean dryRun, boolean suppressLayouts,
            boolean failFast);

    /**
     * Sets the client checksum on the target file. Fails if the target doesn't exist or not a file.
     *
     * @param repo               The repository holding the file
     * @param checksumType       The {@link org.artifactory.checksum.ChecksumType}
     * @param targetFileRepoPath Target file
     * @param checksum           Checksum value
     * @return Checksum info for the given type
     */
    @Lock
    ChecksumInfo setClientChecksum(LocalRepo repo, ChecksumType checksumType, RepoPath targetFileRepoPath,
            String checksum);

    /**
     * Undeploy (deletes) the given path in a transaction.
     *
     * @param repoPath The path to delete
     */
    @Lock
    BasicStatusHolder undeployInTx(RepoPath repoPath, boolean calcMavenMetadata, BasicStatusHolder statusHolder);

    @Lock
    RepositoriesCache rebuildRepositoriesInTransaction();

    //TODO: [by YS] cannot be transactional by itself should be transactional mandatory
    MutableVfsItem getMutableItem(RepoPath repoPath);

    /**
     * Updates the {@param path}'s row in the nodes table with {@param sha2}
     * BEWARE - this method is meant for a very specific usage - the sha2 db conversion, and for that reason i've made
     * it as un-generic as I could to stop you from using it for anything else.
     * If you think it's a good idea to change a node's checksum - stop.
     * you have no business changing the checksums calculated and verified by the Binarystore - it will always lead to
     * inconsistencies and security issues!
     */
    @Async(workQueue = true)
    void updateSha2(ChecksumCalculationWorkItem workItem);

    /**
     * Updates the {@param path}'s row in the nodes table with {@param sha2}
     * BEWARE - this method is meant for a very specific usage - the sha2 db conversion, and for that reason i've made
     * it as un-generic as I could to stop you from using it for anything else.
     * If you think it's a good idea to change a node's checksum - stop.
     * you have no business changing the checksums calculated and verified by the Binarystore - it will always lead to
     * inconsistencies and security issues!
     */
    @Async(transactional = true, workQueue = true)
    void updateRepoPathChecksum(RepoPathChecksumCalculationWorkItem workItem);

    /**
     * BEWARE - this method is meant for a very specific usage - the sha2 db conversion --> see above.
     * INTERNAL USE ONLY
     */
    @Lock
    void updateSha2ForPath(RepoPath path, String sha2);

    List<String> getAllLocalRepoKeys();

    /**
     * Retrieve the default deployment repo key for {@param localOrVirtualRepoKey} in case of virtual.
     * In case the given repo key is local it is returned.
     *
    **/
    String getDefaultDeploymentRepoKey(String localOrVirtualRepoKey);
}