package org.artifactory.storage.service;

import com.google.common.collect.ImmutableMap;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.binstore.service.BinariesGarbageCollectorService;
import org.artifactory.storage.binstore.service.BinaryService;
import org.artifactory.storage.binstore.service.GCProviderWrapper;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.artifactory.storage.db.binstore.service.garbage.BinariesGCProvider;
import org.artifactory.storage.db.binstore.service.garbage.GCCandidateProducer;
import org.artifactory.storage.db.binstore.service.garbage.TrashAndBinariesGCProvider;
import org.artifactory.storage.db.binstore.service.garbage.TrashGCProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.Collections;

/**
 * @author Uriah Levy
 */
@Service
public class BinariesGarbageCollectorServiceImpl implements BinariesGarbageCollectorService {

    private static final Logger log = LoggerFactory.getLogger(BinariesGarbageCollectorServiceImpl.class);

    private RepositoryService repoService;
    private TrashService trashService;
    private BinaryService binaryService;
    private SecurityService securityService;
    private ImmutableMap<GarbageCollectorStrategy, GCCandidateProducer> gcProducers;

    @Autowired
    public BinariesGarbageCollectorServiceImpl(RepositoryService repoService, TrashService trashService,
            BinaryService binaryService, SecurityService securityService) {
        this.repoService = repoService;
        this.trashService = trashService;
        this.binaryService = binaryService;
        this.securityService = securityService;
        initProducers();
    }

    @Override
    public void startGCByStrategy(GarbageCollectorStrategy strategy) {
        GCCandidateProducer gcCandidateProducer = gcProducers.get(strategy);
        if (!gcCandidateProducer.isProducerRunning()) {
            gcCandidateProducer.startProducer();
        } else {
            log.warn("GC strategy ({}) is already running", strategy.name().toLowerCase());
        }
    }

    @Override
    public void stopGCByStrategy(GarbageCollectorStrategy strategy) {
        GCCandidateProducer gcCandidateProducer = gcProducers.get(strategy);
        gcCandidateProducer.askToStopProducer();
        log.info("GC strategy ({}) scheduled to stop", strategy.name().toLowerCase());
    }

    private void initProducers() {
        gcProducers = ImmutableMap.of(GarbageCollectorStrategy.FULL, getFullGCStrategyProducer(),
                GarbageCollectorStrategy.TRASH, getTrashGCCandidateProducer());
    }

    private GCCandidateProducer getTrashGCCandidateProducer() {
        return new GCCandidateProducer(
                Collections.singletonList(
                        new GCProviderWrapper(new TrashAndBinariesGCProvider(repoService, trashService, binaryService))),
                GarbageCollectorStrategy.TRASH, securityService);
    }

    private GCCandidateProducer getFullGCStrategyProducer() {
        return new GCCandidateProducer(Arrays.asList(new GCProviderWrapper(new TrashGCProvider(trashService)),
                new BinariesGCProvider(binaryService, repoService)), GarbageCollectorStrategy.FULL, securityService);
    }
}
