/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo;

import org.apache.commons.io.IOUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.LayoutsCoreAddon;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.StatusHolder;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.io.checksum.ChecksumUtils;
import org.artifactory.repo.db.DbCacheRepo;
import org.artifactory.repo.db.DbStoringRepoMixin;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.util.RepoLayoutUtils;
import org.mockito.Mockito;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.InputStream;

/**
 * This class tests the behaviour of the allowsDownloadMethod and checks that it returns the proper results on different
 * scenarios
 *
 * @author Noam Tenne
 */
public class RemoteRepoBaseTest extends ArtifactoryHomeBoundTest {
    private AuthorizationService authService = Mockito.mock(AuthorizationService.class);
    private InternalRepositoryService internalRepoService = Mockito.mock(InternalRepositoryService.class);
    private InternalArtifactoryContext context = Mockito.mock(InternalArtifactoryContext.class);
    private AddonsManager addonsManager = Mockito.mock(AddonsManager.class);
    private LayoutsCoreAddon layoutsCoreAddon = Mockito.mock(LayoutsCoreAddon.class);
    private HaAddon haAddon = Mockito.mock(HaAddon.class);
    private XrayAddon xrayAddon = Mockito.mock(XrayAddon.class);

    private HttpRepoDescriptor httpRepoDescriptor = new HttpRepoDescriptor();
    private HttpRepo httpRepo;
    private RepoPath repoPath = InternalRepoPathFactory.create("remote-repo-cache", "test/test/1.0/test-1.0.jar");

    @BeforeMethod
    public void setup() {
        Mockito.when(context.getAuthorizationService()).thenReturn(authService);
        Mockito.when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        Mockito.when(addonsManager.addonByType(LayoutsCoreAddon.class)).thenReturn(layoutsCoreAddon);
        Mockito.when(addonsManager.addonByType(HaAddon.class)).thenReturn(haAddon);
        Mockito.when(addonsManager.addonByType(XrayAddon.class)).thenReturn(xrayAddon);
        ArtifactoryContextThreadBinder.bind(context);
        httpRepoDescriptor.setRepoLayout(RepoLayoutUtils.MAVEN_2_DEFAULT);
        httpRepo = new HttpRepo(httpRepoDescriptor, internalRepoService, false, null);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    /**
     * Try to download with ordinary path
     */
    @Test
    public void testDownloadWithCachePrefix() {
        expectCanRead(true);
        expectXrayBlocked();
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertFalse(statusHolder.isError());
        verifyCanRead();
    }

    /**
     * Try to download while supplying a path with no "-cache" prefix
     */
    @Test
    public void testDownloadWithNoCachePrefix() {
        RepoPath repoPathNoPrefix = InternalRepoPathFactory.create("remote-repo", "test/test/1.0/test-1.0.jar");
        expectCanRead(true);
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPathNoPrefix);
        Assert.assertFalse(statusHolder.isError());
        verifyCanRead();
    }

    /**
     * Try to download when Anonymous mode is enabled
     */
    @Test
    public void testDownloadWithAnonymousEnabledAndNoReadPermissions() {
        expectCanRead(false);       // don't give read access
        expectXrayBlocked();
        createCacheRepo();   // and enable anon
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertTrue(statusHolder.isError());
        verifyCanRead();
    }

    private void verifyCanRead() {
        Mockito.verify(authService).canRead(repoPath);
        Mockito.reset(authService);
    }

    /**
     * Try to download when the repo is blacked out
     */
    @Test
    public void testDownloadWithBlackedOut() {
        httpRepoDescriptor.setBlackedOut(true);
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertTrue(statusHolder.isError());
        httpRepoDescriptor.setBlackedOut(false);
    }

    /**
     * Try to download when the repo does not serve releases
     */
    @Test
    public void testDownloadWithNoReleases() {
        httpRepoDescriptor.setHandleReleases(false);
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertTrue(statusHolder.isError());
        httpRepoDescriptor.setHandleReleases(true);
    }

    /**
     * Try to download when requested resource is excluded
     */
    @Test
    public void testDownloadWithExcludes() {
        httpRepoDescriptor.setExcludesPattern("test/test/1.0/**");
        createCacheRepo();
        StatusHolder statusHolder = tryAllowsDownload(repoPath);
        Assert.assertTrue(statusHolder.isError());
        httpRepoDescriptor.setExcludesPattern("");
    }

    /**
     * Verify reading of valid checksum file to test the {@link ChecksumUtils#checksumStringFromStream(InputStream)}
     * method
     */
    @Test
    public void testReadValidChecksum() throws Exception {
        invokeReadChecksum("valid", "dasfasdf4r234234q32asdfadfasasdfasdf");
    }

    /**
     * Verify reading of a checksum file containing comments to test the {@link ChecksumUtils#checksumStringFromStream(InputStream)}
     * method
     */
    @Test
    public void testReadCommentChecksum() throws Exception {
        invokeReadChecksum("comment", "asdfaeaef435345435asdf");
    }

    /**
     * Verify reading of a checksum file containing file description to test the {@link
     * ChecksumUtils#checksumStringFromStream(InputStream)} method
     */
    @Test
    public void testReadDescChecksum() throws Exception {
        invokeReadChecksum("desc", "dasfasdf4r234234q32asdfadfasdf");
    }

    /**
     * Verify reading of an empty checksum file to test the {@link ChecksumUtils#checksumStringFromStream(InputStream)}
     * method
     */
    @Test
    public void testReadEmptyChecksum() throws Exception {
        invokeReadChecksum("empty", "");
    }

    /**
     * Check the different values that are read from the given checksum are as expected
     *
     * @param checksumType  Type of checksum test file to read
     * @param expectedValue Expected checksum value
     */
    private void invokeReadChecksum(String checksumType, String expectedValue) throws Exception {
        InputStream stream = getClass().getResourceAsStream("/org/artifactory/repo/test-" + checksumType + ".md5");
        try {
            String checksum = ChecksumUtils.checksumStringFromStream(stream);
            Assert.assertEquals(expectedValue, checksum, "Incorrect checksum value returned.");
        } finally {
            IOUtils.closeQuietly(stream);
        }
    }

    private void createCacheRepo() {
        httpRepoDescriptor.setKey("remote-repo");
        DbCacheRepo localCacheRepo = new DbCacheRepo(httpRepo, null);
        ReflectionTestUtils.setField(httpRepo, "localCacheRepo", localCacheRepo);

        DbStoringRepoMixin mixin = Mockito.mock(DbStoringRepoMixin.class);
        ReflectionTestUtils.setField(localCacheRepo, "mixin", mixin);
        Mockito.when(mixin.itemExists(Mockito.any())).thenReturn(true);
        Mockito.when(mixin.getKey()).thenReturn("remote-repo-cache");
    }

    private void expectCanRead(boolean canRead) {
        Mockito.when(authService.canRead(repoPath)).thenReturn(canRead);
    }

    private void expectXrayBlocked() {
        Mockito.when(xrayAddon.isDownloadBlocked(Mockito.any(RepoPath.class))).thenReturn(false);
    }

    /**
     * Try to download the the given path
     *
     * @param path Repopath to test the download with
     * @return StatusHolder - Contains any errors which might be monitored
     */
    private StatusHolder tryAllowsDownload(RepoPath path) {
        return httpRepo.checkDownloadIsAllowed(path);
    }
}
