/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.api.release.bundle;

import lombok.NonNull;
import org.artifactory.api.component.ComponentDetails;
import org.artifactory.api.repo.Async;
import org.artifactory.api.rest.distribution.bundle.models.ReleaseBundleModel;
import org.artifactory.bundle.LightArtifactsBundle;
import org.artifactory.repo.RepoPath;
import org.artifactory.sapi.common.Lock;
import org.jfrog.storage.StorageException;

import javax.annotation.Nonnull;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;

/**
 * @author Rotem Kfir
 */
public interface ReleaseBundleService {
    /**
     * Find all Release Bundles with incomplete transaction that has been created over a predefined time period ago
     *
     * @return a list of Release Bundles ids
     *
     * @throws StorageException In case a storage error occurred during the operation
     */
    @Nonnull
    List<Long> findExpiredBundlesIds(Long cleanupPeriod) throws StorageException;

    /**
     * Deletes a Release Bundle either with a closed or incomplete transaction by its id, along with its files and blob
     *
     * @param id            the id by which to delete the Release Bundle
     * @param deleteContent if true then artifacts related to this bundle will be deleted from the repositories
     * @return true if a Release Bundle was deleted, false if the Release Bundle was not found
     *
     * @throws StorageException In case a storage error occurred during the operation
     */
    @Lock
    boolean deleteBundle(Long id, boolean deleteContent) throws StorageException;

    /**
     * Redistributes a Release Bundle with an incomplete transaction by its id
     *
     * @param id the id by which to find the Release Bundle
     * @return true if a Release Bundle was updated, false if the Release Bundle was not found or is not incomplete
     *
     * @throws StorageException In case a storage error occurred during the operation
     */
    boolean redistributeBundle(Long id) throws StorageException;

    /**
     * Checks if the given repo path is a part of a Release Bundle
     *
     * @param path the repo path to check
     * @return a list of the bundles the given path is related to (either the path is a file that is part of a bundle or
     * a folder that contains file/s that are part of a bundle)
     *
     * @throws StorageException In case a storage error occurred during the operation
     */
    @Nonnull
    List<LightArtifactsBundle> getBundlesRelatedToRepoPath(String path) throws StorageException;

    void deleteAllBundles();

    @Async(delayUntilAfterCommit = true, authenticateAsSystem = true, transactional = true)
    void copyBundleArtifactsAsync(ReleaseBundleModel releaseBundleModel, String storingRepo,
            Map<String, String> artifactMapping)
            throws SQLException;

    void createBundleNode(long artifactsBundleId, RepoPath targetRepoPathm, ComponentDetails componentDetails)
            throws SQLException;

    /**
     * Get previously calculated component details of a (possibly deleted) artifact
     *
     * @param repoPath - path of the (possibly non-existent) node
     * @return object containing details required for smart replication
     */
    ComponentDetails getOriginalComponentDetails(@NonNull RepoPath repoPath) throws SQLException;
}
