/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.bundle.itest.dao;

import org.artifactory.bundle.BundleTransactionStatus;
import org.artifactory.bundle.BundleType;
import org.artifactory.storage.db.binstore.dao.BinariesDao;
import org.artifactory.storage.db.binstore.entity.BinaryEntity;
import org.artifactory.storage.db.bundle.dao.ArtifactBundlesDao;
import org.artifactory.storage.db.bundle.model.BundleNode;
import org.artifactory.storage.db.bundle.model.DBArtifactsBundle;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.entity.Node;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.joda.time.DateTime;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.List;

import static org.fest.assertions.Assertions.assertThat;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNull;

/**
 * @author Rotem Kfir
 */
@Test
public class ArtifactBundlesDaoTest extends DbBaseTest {

    @Autowired
    protected ArtifactBundlesDao artifactBundlesDao;

    @Autowired
    protected NodesDao nodesDao;

    @Autowired
    protected BinariesDao binariesDao;

    private long nextId = 1;

    public void test() throws SQLException {
        List<DBArtifactsBundle> bundles = artifactBundlesDao.getCompletedBundlesLastVersion(BundleType.TARGET);
        assertThat(bundles).hasSize(0);

        DBArtifactsBundle bundle = new DBArtifactsBundle();
        for (int i = 0; i < 3; i++) {
            bundle.setId(++nextId);
            bundle.setName("RB");
            bundle.setStatus(BundleTransactionStatus.COMPLETE);
            bundle.setDateCreated(DateTime.now().plusMillis((int) nextId));
            bundle.setVersion(String.valueOf(nextId));
            bundle.setType(BundleType.TARGET);
            bundle.setSignature("Signature" + nextId);
            artifactBundlesDao.create(bundle);
        }

        // Same creation date, higher version
        bundle.setId(++nextId);
        bundle.setVersion(String.valueOf(nextId));
        artifactBundlesDao.create(bundle);
        // Same creation date, even higher version, but INPROGRESS
        bundle.setId(++nextId);
        bundle.setVersion(String.valueOf(nextId));
        bundle.setStatus(BundleTransactionStatus.INPROGRESS);
        artifactBundlesDao.create(bundle);

        bundles = artifactBundlesDao.getCompletedBundlesLastVersion(BundleType.TARGET);
        assertThat(bundles).hasSize(1);
        assertThat(bundles.get(0).getVersion()).isEqualTo(String.valueOf(nextId - 1)); // the completed bundle with the latest date_created and highest version name

        bundle.setId(++nextId);
        bundle.setName("RB2");
        bundle.setStatus(BundleTransactionStatus.COMPLETE);
        artifactBundlesDao.create(bundle);
        bundles = artifactBundlesDao.getCompletedBundlesLastVersion(BundleType.TARGET);
        assertThat(bundles).hasSize(2);
    }

    public void testCreateBundleFile() throws SQLException {
        DBArtifactsBundle dbArtifactsBundle = new DBArtifactsBundle();
        dbArtifactsBundle.setId(100);
        dbArtifactsBundle.setStatus(BundleTransactionStatus.INPROGRESS);
        dbArtifactsBundle.setName("yoazRB");
        dbArtifactsBundle.setDateCreated(new DateTime(System.currentTimeMillis()));
        dbArtifactsBundle.setSignature("MySig");
        dbArtifactsBundle.setStoringRepo("release-bundles");
        dbArtifactsBundle.setType(BundleType.SOURCE);
        dbArtifactsBundle.setVersion("1.x.0");
        artifactBundlesDao.create(dbArtifactsBundle);
        BinaryEntity binaryEntity = new BinaryEntity("mySha1", "mySha2", "myMd5", 100);
        binariesDao.create(binaryEntity);
        Node node = new Node(3, true, "repo2", "path/to", "file", (short) 2, 543L, "yossis",
                123L, "modifier", 213L, 22, "mySha1",
                "mySha1", "myMd5", "myMd5", "mySha2");
        nodesDao.create(node);
        String componentDetails = "{\"name\":\"yoaz.jar\",\"version\":\"1.0.x\",\"componentType\":\"Maven\",\"extension\":\"jar\",\"mimeType\":\"application/java-archive\"}";
        BundleNode bundleNode = createBundleNode(node.getNodeId(), dbArtifactsBundle.getId(), componentDetails, "my/repo/path", 1231);
        artifactBundlesDao.create(bundleNode);
        bundleNode = createBundleNode(node.getNodeId(), dbArtifactsBundle.getId(), null, "my/repo/path2", 1232);
        artifactBundlesDao.create(bundleNode);
        String details = artifactBundlesDao.getComponentDetails("my/repo/path");
        assertEquals(componentDetails, details);
        details = artifactBundlesDao.getComponentDetails("my/repo/path2");
        assertNull(details);
    }

    private BundleNode createBundleNode(long nodeID, long bundleID, String componentDetails, String bundleNodeRepoPath, int bundleNodeID) {
        BundleNode bundleNode = new BundleNode();
        bundleNode.setBundleId(bundleID);
        bundleNode.setId(bundleNodeID);
        bundleNode.setOriginalFileDetails(componentDetails);
        bundleNode.setRepoPath(bundleNodeRepoPath);
        bundleNode.setNodeId(nodeID);
        return bundleNode;
    }

    public void testSource() throws SQLException {
        DBArtifactsBundle bundle = new DBArtifactsBundle();
        bundle.setId(++nextId);
        bundle.setName("RBS");
        bundle.setStatus(BundleTransactionStatus.INPROGRESS);
        bundle.setDateCreated(DateTime.now());
        bundle.setVersion("1.0");
        bundle.setType(BundleType.SOURCE);
        bundle.setSignature("Signature");
        artifactBundlesDao.create(bundle);

        DBArtifactsBundle rb = artifactBundlesDao.getArtifactsBundle("RBS", "1.0", BundleType.SOURCE);
        assertThat(rb).isNotNull();
        artifactBundlesDao.failBundle("RBS", "1.0", BundleType.SOURCE.name());
        rb = artifactBundlesDao.getArtifactsBundle("RBS", "1.0", BundleType.SOURCE);
        assertThat(rb).isNotNull();
        assertThat(rb.getStatus()).isEqualTo(BundleTransactionStatus.FAILED);

    }
}
