/*
 *
 * Copyright 2016 JFrog Ltd. All rights reserved.
 * JFROG PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package org.artifactory.rest.common.security;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import org.artifactory.rest.common.util.RestUtils;
import org.artifactory.security.*;
import org.artifactory.security.permissions.PermissionTargetModel;
import org.artifactory.security.permissions.RepoPermissionTargetModel;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.security.PermissionTargetNaming.NAMING_BACKEND;

/**
 * @author Noam Y. Tenne
 */
public abstract class SecurityModelPopulator {

    private SecurityModelPopulator() {

    }

    @Nonnull
    public static UserConfigurationImpl getUserConfiguration(@Nonnull UserInfo user) {
        UserConfigurationImpl userConfiguration = new UserConfigurationImpl();
        long lastLoginTimeMillis = user.getLastLoginTimeMillis();
        if (lastLoginTimeMillis > 0) {
            userConfiguration.setLastLoggedIn(RestUtils.toIsoDateString(lastLoginTimeMillis));
        }
        userConfiguration.setRealm(user.getRealm());
        userConfiguration.setAdmin(user.isAdmin());
        userConfiguration.setEmail(user.getEmail());
        userConfiguration.setName(user.getUsername());
        userConfiguration.setProfileUpdatable(user.isUpdatableProfile());
        userConfiguration.setInternalPasswordDisabled(user.isPasswordDisabled());
        Set<UserGroupInfo> groups = user.getGroups();
        if ((groups != null) && !groups.isEmpty()) {
            userConfiguration.setGroups(Sets.newHashSet(groups.stream().map(input -> {
                        if (input == null) {
                            return null;
                        }
                        return input.getGroupName();
                    }).collect(Collectors.toList())
            ));
        }
        Set<UserPropertyInfo> properties = user.getUserProperties();
        List<UserPropertyInfo> disableUiProperty = properties.stream()
                .filter(userPropertyInfo -> userPropertyInfo.getPropKey().equals("blockUiView"))
                .collect(Collectors.toList());
        if (!disableUiProperty.isEmpty()) {
            userConfiguration.setdisableUIAccess(Boolean.valueOf(disableUiProperty.get(0).getPropValue()));
        }
        return userConfiguration;
    }

    @Nonnull
    public static GroupConfigurationImpl getGroupConfiguration(@Nonnull GroupInfo group) {
        GroupConfigurationImpl groupConfiguration = new GroupConfigurationImpl();
        groupConfiguration.setDescription(group.getDescription());
        groupConfiguration.setAutoJoin(group.isNewUserDefault());
        groupConfiguration.setName(group.getGroupName());
        groupConfiguration.setRealm(group.getRealm());
        groupConfiguration.setRealmAttributes(group.getRealmAttributes());
        groupConfiguration.setAdminPrivileges(group.isAdminPrivileges());
        return groupConfiguration;
    }

    @Nonnull
    public static PermissionTargetConfigurationImpl getPermissionTargetConfiguration(@Nonnull RepoAcl acl) {
        PermissionTargetConfigurationImpl permissionTargetConfiguration = new PermissionTargetConfigurationImpl();
        RepoPermissionTarget permissionTarget = acl.getPermissionTarget();
        permissionTargetConfiguration.setName(permissionTarget.getName());
        permissionTargetConfiguration.setIncludesPattern(permissionTarget.getIncludesPattern());
        permissionTargetConfiguration.setExcludesPattern(permissionTarget.getExcludesPattern());
        permissionTargetConfiguration.setRepositories(permissionTarget.getRepoKeys());

        PrincipalConfiguration principalConfiguration = getPrincipalConfiguration(acl.getMutableAces(), NAMING_BACKEND);
        permissionTargetConfiguration.setPrincipals(principalConfiguration);
        return permissionTargetConfiguration;
    }

    @Nonnull
    public static PermissionTargetModel getPermissionTargetModelV2(@Nullable RepoAcl repoAcl, @Nullable BuildAcl buildAcl, PermissionTargetNaming outputMode) {
        PermissionTargetModel permissionTargetModel = new PermissionTargetModel();
        populateRepoPermissionTargetModel(repoAcl, permissionTargetModel, outputMode);
        populateBuildPermissionTargetModel(buildAcl, permissionTargetModel, outputMode);
        return permissionTargetModel;
    }

    private static void populateRepoPermissionTargetModel(@Nullable RepoAcl repoAcl, PermissionTargetModel permissionTargetModel, PermissionTargetNaming outputMode) {
        if (repoAcl != null) {
            RepoPermissionTarget repoPermissionTarget = repoAcl.getPermissionTarget();
            PrincipalConfiguration principalConfiguration = getPrincipalConfiguration(repoAcl.getMutableAces(), outputMode);
            RepoPermissionTargetModel repoPermissionTargetModel =
                    RepoPermissionTargetModel.builder()
                    .includePatterns(repoPermissionTarget.getIncludes())
                    .excludePatterns(repoPermissionTarget.getExcludes())
                    .repositories(repoPermissionTarget.getRepoKeys())
                    .actions(principalConfiguration).build();
            permissionTargetModel.setName(repoPermissionTarget.getName());
            permissionTargetModel.setRepo(repoPermissionTargetModel);
        }
    }

    private static void populateBuildPermissionTargetModel(@Nullable BuildAcl buildAcl, PermissionTargetModel permissionTargetModel, PermissionTargetNaming outputMode) {
        if (buildAcl != null) {
            BuildPermissionTarget buildPermissionTarget = buildAcl.getPermissionTarget();
            PrincipalConfiguration principalConfiguration = getPrincipalConfiguration(buildAcl.getMutableAces(), outputMode);
            RepoPermissionTargetModel buildPermissionTargetModel = RepoPermissionTargetModel.builder()
                    .includePatterns(buildPermissionTarget.getIncludes())
                    .excludePatterns(buildPermissionTarget.getExcludes())
                    .repositories(buildPermissionTarget.getRepoKeys())
                    .actions(principalConfiguration)
                    .build();
            permissionTargetModel.setName(buildPermissionTarget.getName());
            permissionTargetModel.setBuild(buildPermissionTargetModel);
        }
    }

    /**
     * {@param outputMode} used for v2 api and ui, where we output human-readable permission strings, or camel cased.
     * (internally they are still saved as short strings - see {@link ArtifactoryPermission})
     */
    private static PrincipalConfiguration getPrincipalConfiguration(Set<MutableAceInfo> aces, PermissionTargetNaming outputMode) {
        Map<String, Set<String>> users = Maps.newHashMap();
        Map<String, Set<String>> groups = Maps.newHashMap();

        for (AceInfo ace : aces) {
            String principal = ace.getPrincipal();
            Set<String> permissionsAsString;
            permissionsAsString = getPermissionsByNamingConvention(ace, outputMode);
            if (ace.isGroup()) {
                groups.put(principal, permissionsAsString);
            } else {
                users.put(principal, permissionsAsString);
            }
        }
        PrincipalConfiguration principalConfiguration = new PrincipalConfiguration();
        if (!users.isEmpty()) {
            principalConfiguration.setUsers(users);
        }
        if (!groups.isEmpty()) {
            principalConfiguration.setGroups(groups);
        }
        return principalConfiguration;
    }

    private static Set<String> getPermissionsByNamingConvention(AceInfo ace, PermissionTargetNaming outputMode) {
        Set<String> permissionsAsString;
        switch (outputMode) {
            case NAMING_BACKEND: //v1 api and access
                permissionsAsString = ace.getPermissionsAsString();
                break;
            case NAMING_DISPLAY:
                permissionsAsString = ace.getPermissionsDisplayNames();
                break;
            case NAMING_UI:
                permissionsAsString = ace.getPermissionsUiNames();
                break;
            default:
                permissionsAsString = ace.getPermissionsAsString();
                break;
        }
        return permissionsAsString;
    }

    public static Set<String> getPermissionsAsString(boolean canRead, boolean canAnnotate, boolean canDeploy,
            boolean canDelete, boolean canAdmin) {

        Set<String> permissionsAsString = Sets.newHashSet();
        if (canRead) {
            appendPermissionString(permissionsAsString, ArtifactoryPermission.READ);
        }
        if (canAnnotate) {
            appendPermissionString(permissionsAsString, ArtifactoryPermission.ANNOTATE);
        }
        if (canDeploy) {
            appendPermissionString(permissionsAsString, ArtifactoryPermission.DEPLOY);
        }
        if (canDelete) {
            appendPermissionString(permissionsAsString, ArtifactoryPermission.DELETE);
        }
        if (canAdmin) {
            appendPermissionString(permissionsAsString, ArtifactoryPermission.MANAGE);
        }
        return permissionsAsString;
    }

    private static void appendPermissionString(Set<String> permissionsAsString, ArtifactoryPermission permission) {
        permissionsAsString.add(permission.getString());
    }
}
