package org.artifactory.rest.services.permissions;

import com.google.common.collect.Sets;
import org.apache.commons.lang3.tuple.Pair;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.security.AclService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.model.xstream.security.AceImpl;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.security.*;

import javax.annotation.Nonnull;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.BiConsumer;

import static org.artifactory.security.PermissionTarget.*;
import static org.artifactory.util.CollectionUtils.notNullOrEmpty;

/**
 * (Bunch of shared code taken out from {@see RestSecurityRequestHandler})
 * @author Noam Y. Tenne
 * @author Dan Feldman
 */
public class RestSecurityHelperCommon {

    private static final String ERR_CANNOT_ADD_ADMIN = "' has admin privileges, and cannot be added to a Permission Target.";
    private static final String ERR_NON_EXISTING_PRINCIPAL = "Permission target contains a reference to a non-existing ";
    public static final String CONFLICT_ERR_MSG = "The permission target name that was " +
            "provided in the request path does not match the permission name in the provided permission " +
            "configuration object.";

    private RestSecurityHelperCommon() {}

    /**
     * @param entityKey Permission target to look for
     * @return (Repo, Build)
     */
    public static Pair<RepoAcl, BuildAcl> getExistingPermissionTarget(AclService aclService, String entityKey) {
        return Pair.of(aclService.getRepoAcl(entityKey), aclService.getBuildAcl(entityKey));
    }

    /**
     * @param inputMode governs which naming convention should be used when parsing permission action strings
     */
    @Nonnull
    public static Set<AceInfo> getAcesForAcl(PrincipalConfiguration actions, PermissionTargetNaming inputMode) {
        if (actions != null) {
            Set<AceInfo> aces = Sets.newHashSet();
            addPrincipalAces(aces, actions.getUsers(), false, inputMode);
            addPrincipalAces(aces, actions.getGroups(), true, inputMode);
            return aces;
        }
        return Sets.newHashSet();
    }

    /**
     * @param inputMode governs which naming convention should be used when parsing permission action strings
     */
    private static void addPrincipalAces(Set<AceInfo> aces, Map<String, Set<String>> action, boolean group, PermissionTargetNaming inputMode) {
        if (action != null) {
            for (Map.Entry<String, Set<String>> principalToActions : action.entrySet()) {
                AceImpl ace = new AceImpl();
                ace.setPrincipal(principalToActions.getKey());
                setPermissionActionsByNamingConvention(inputMode, ace, principalToActions.getValue());
                ace.setGroup(group);
                aces.add(ace);
            }
        }
    }

    private static void setPermissionActionsByNamingConvention(PermissionTargetNaming inputMode, AceImpl ace,
            Set<String> actions) {
        switch (inputMode) {
            case NAMING_BACKEND:
                ace.setPermissionsFromStrings(actions);
                break;
            case NAMING_DISPLAY:
                ace.setPermissionsFromDisplayNames(actions);
                break;
            case NAMING_UI:
                ace.setPermissionsFromUiNames(actions);
                break;
            default:
                ace.setPermissionsFromStrings(actions);
                break;
        }
    }

    /**
     * @return the first non-existing repo contained in the {@param repositories} list
     */
    public static String getFirstNonExistingRepoFromList(List<String> repositories, CentralConfigService configService, AclService aclService) {
        if (repositories != null) {
            MutableCentralConfigDescriptor mutableDescriptor = configService.getMutableDescriptor();
            //remote-repo-cache should be handled like remote-repo, so remove the '-cache' suffix
            for (String repository : aclService.convertCachedRepoKeysToRemote(repositories)) {
                if (!mutableDescriptor.isRepositoryExists(repository)
                        && !ANY_REPO.equals(repository)
                        && !ANY_REMOTE_REPO.equals(repository)
                        && !ANY_LOCAL_REPO.equals(repository)
                        && !ANY_DISTRIBUTION_REPO.equals(repository)) {
                    return repository;
                }
            }
        }
        return null;
    }

    /**
     * @throws BadRequestException on non-existing principals
     */
    public static void checkForNonExistingPrinciples(UserGroupService userGroupService, PrincipalConfiguration principals) throws BadRequestException {
        if (principals != null) {
            checkForNonExistingPrincipals(userGroupService, principals.getUsers(), RestSecurityHelperCommon::validateNonExitingUser);
            checkForNonExistingPrincipals(userGroupService, principals.getGroups(), RestSecurityHelperCommon::validateNonExistingGroup);
        }
    }

    private static void checkForNonExistingPrincipals(UserGroupService userGroupService,
            Map<String, Set<String>> principals, BiConsumer<UserGroupService, String> validator) throws BadRequestException {
        if (principals != null) {
            principals.entrySet().stream()
                    .filter(principalEntry -> notNullOrEmpty(principalEntry.getValue()))
                    .map(Map.Entry::getKey)
                    .forEach(principalName -> validator.accept(userGroupService, principalName));
        }
    }

    private static void validateNonExitingUser(UserGroupService userGroupService, String username) {
        try {
            UserInfo user = userGroupService.findUser(username);
            if (user.isEffectiveAdmin()) {
                throw new BadRequestException("User '" + username + ERR_CANNOT_ADD_ADMIN);
            }
        } catch (Exception e) {
            throw new BadRequestException(ERR_NON_EXISTING_PRINCIPAL + "user '" + username + "'.");
        }
    }

    private static void validateNonExistingGroup(UserGroupService userGroupService, String groupName) {
        GroupInfo group = userGroupService.findGroup(groupName);
        if (group == null) {
            throw new BadRequestException(ERR_NON_EXISTING_PRINCIPAL + "group '" + groupName + "'.");
        }
        if (group.isAdminPrivileges()) {
            throw new BadRequestException("Group '" + groupName + ERR_CANNOT_ADD_ADMIN);
        }
    }
}
